@@ -1,5 +1,4 @@
 /*!
-
 *xmlparser* is a low-level, pull-based, zero-allocation
 [XML 1.0](https://www.w3.org/TR/xml/) parser.
 
@@ -11,43 +10,42 @@ for token in xmlparser::Tokenizer::from("<tagname name='value'/>") {
 }
 ```
 
-## Why a new library
+## Why a new library?
 
-This library is basically a low-level XML tokenizer that preserves a position of the tokens
-and does not intend to be used directly.
-If you are looking for a more high-level solution - checkout
+This library is basically a low-level XML tokenizer that preserves the positions of the tokens
+and is not intended to be used directly.
+If you are looking for a higher level solution, check out
 [roxmltree](https://github.com/RazrFalcon/roxmltree).
 
 ## Benefits
 
-- All tokens contain `StrSpan` objects which contain a position of the data in the original document.
-- Good error processing. All error types contain position (line:column) where it occurred.
+- All tokens contain `StrSpan` structs which represent the position of the substring
+  in the original document.
+- Good error processing. All error types contain the position (line:column) where it occurred.
 - No heap allocations.
 - No dependencies.
-- Tiny. ~1500 LOC and ~40KiB in the release build according to the `cargo-bloat`.
+- Tiny. ~1400 LOC and ~30KiB in the release build according to `cargo-bloat`.
+- Supports `no_std` builds. To use without the standard library, disable the default features.
 
 ## Limitations
 
-- Currently, only ENTITY objects are parsed from the DOCTYPE. Other ignored.
+- Currently, only ENTITY objects are parsed from the DOCTYPE. All others are ignored.
 - No tree structure validation. So an XML like `<root><child></root></child>`
   or a string without root element
   will be parsed without errors. You should check for this manually.
   On the other hand `<a/><a/>` will lead to an error.
-- Duplicated attributes is not an error. So an XML like `<item a="v1" a="v2"/>`
+- Duplicated attributes is not an error. So XML like `<item a="v1" a="v2"/>`
   will be parsed without errors. You should check for this manually.
 - UTF-8 only.
 
 ## Safety
 
-- The library must not panic. Any panic considered as a critical bug
+- The library must not panic. Any panic is considered a critical bug
   and should be reported.
-- The library forbids the unsafe code.
+- The library forbids unsafe code.
 */
 
 #![no_std]
-#![cfg_attr(feature = "cargo-clippy", allow(unreadable_literal))]
-
-#![doc(html_root_url = "https://docs.rs/xmlparser/0.11.0")]
 
 #![forbid(unsafe_code)]
 #![warn(missing_docs)]
@@ -58,9 +56,6 @@ If you are looking for a more high-level solution - checkout
 extern crate std;
 
 
-use core::fmt;
-
-
 macro_rules! matches {
     ($expression:expr, $($pattern:tt)+) => {
         match $expression {
@@ -76,15 +71,15 @@ mod stream;
 mod strspan;
 mod xmlchar;
 
-pub use error::*;
-pub use stream::*;
-pub use strspan::*;
-pub use xmlchar::*;
+pub use crate::error::*;
+pub use crate::stream::*;
+pub use crate::strspan::*;
+pub use crate::xmlchar::*;
 
 
 /// An XML token.
 #[allow(missing_docs)]
-#[derive(Clone, Copy, PartialEq, Debug)]
+#[derive(Clone, Copy, PartialEq, Eq, Hash, Debug)]
 pub enum Token<'a> {
     /// Declaration token.
     ///
@@ -242,6 +237,7 @@ pub enum Token<'a> {
     ///
     /// Contains text between elements including whitespaces.
     /// Basically everything between `>` and `<`.
+    /// Except `]]>`, which is not allowed and will lead to an error.
     ///
     /// ```text
     /// <p> text </p>
@@ -268,7 +264,7 @@ pub enum Token<'a> {
 
 
 /// `ElementEnd` token.
-#[derive(Clone, Copy, PartialEq, Debug)]
+#[derive(Clone, Copy, PartialEq, Eq, Hash, Debug)]
 pub enum ElementEnd<'a> {
     /// Indicates `>`
     Open,
@@ -281,7 +277,7 @@ pub enum ElementEnd<'a> {
 
 /// Representation of the [ExternalID](https://www.w3.org/TR/xml/#NT-ExternalID) value.
 #[allow(missing_docs)]
-#[derive(Clone, Copy, PartialEq, Debug)]
+#[derive(Clone, Copy, PartialEq, Eq, Hash, Debug)]
 pub enum ExternalId<'a> {
     System(StrSpan<'a>),
     Public(StrSpan<'a>, StrSpan<'a>),
@@ -290,7 +286,7 @@ pub enum ExternalId<'a> {
 
 /// Representation of the [EntityDef](https://www.w3.org/TR/xml/#NT-EntityDef) value.
 #[allow(missing_docs)]
-#[derive(Clone, Copy, PartialEq, Debug)]
+#[derive(Clone, Copy, PartialEq, Eq, Hash, Debug)]
 pub enum EntityDefinition<'a> {
     EntityValue(StrSpan<'a>),
     ExternalId(ExternalId<'a>),
@@ -301,59 +297,10 @@ type Result<T> = core::result::Result<T, Error>;
 type StreamResult<T> = core::result::Result<T, StreamError>;
 
 
-/// List of token types.
-///
-/// For internal use and errors.
-#[derive(Clone, Copy, PartialEq, Debug)]
-#[allow(missing_docs)]
-pub enum TokenType {
-    XMLDecl,
-    Comment,
-    PI,
-    DoctypeDecl,
-    ElementDecl,
-    AttlistDecl,
-    EntityDecl,
-    NotationDecl,
-    DoctypeEnd,
-    ElementStart,
-    ElementClose,
-    Attribute,
-    CDSect,
-    Whitespace,
-    CharData,
-    Unknown,
-}
-
-impl fmt::Display for TokenType {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        let s = match *self {
-            TokenType::XMLDecl => "Declaration",
-            TokenType::Comment => "Comment",
-            TokenType::PI => "Processing Instruction",
-            TokenType::DoctypeDecl => "Doctype Declaration",
-            TokenType::ElementDecl => "Doctype Element Declaration",
-            TokenType::AttlistDecl => "Doctype Attributes Declaration",
-            TokenType::EntityDecl => "Doctype Entity Declaration",
-            TokenType::NotationDecl => "Doctype Notation Declaration",
-            TokenType::DoctypeEnd => "Doctype End",
-            TokenType::ElementStart => "Element Start",
-            TokenType::ElementClose => "Element Close",
-            TokenType::Attribute => "Attribute",
-            TokenType::CDSect => "CDATA",
-            TokenType::Whitespace => "Whitespace",
-            TokenType::CharData => "Character data",
-            TokenType::Unknown => "Unknown",
-        };
-
-        write!(f, "{}", s)
-    }
-}
-
-
 #[derive(Clone, Copy, PartialEq)]
 enum State {
-    Start,
+    Declaration,
+    AfterDeclaration,
     Dtd,
     AfterDtd,
     Elements,
@@ -374,16 +321,16 @@ pub struct Tokenizer<'a> {
 impl<'a> From<&'a str> for Tokenizer<'a> {
     #[inline]
     fn from(text: &'a str) -> Self {
-        Self::from(StrSpan::from(text))
-    }
-}
+        let mut stream = Stream::from(text);
+
+        // Skip UTF-8 BOM.
+        if stream.starts_with(&[0xEF, 0xBB, 0xBF]) {
+            stream.advance(3);
+        }
 
-impl<'a> From<StrSpan<'a>> for Tokenizer<'a> {
-    #[inline]
-    fn from(span: StrSpan<'a>) -> Self {
         Tokenizer {
-            stream: Stream::from(span),
-            state: State::Start,
+            stream,
+            state: State::Declaration,
             depth: 0,
             fragment_parsing: false,
         }
@@ -391,21 +338,11 @@ impl<'a> From<StrSpan<'a>> for Tokenizer<'a> {
 }
 
 
-/// Shorthand for:
-///
-/// ```no_run
-/// let start = stream.pos() - 2; // or any other number
-/// some_func().map_err(|e|
-///     Error::InvalidToken(Token::SomeToken, stream.gen_error_pos_from(start), Some(e))
-/// )
-/// ```
 macro_rules! map_err_at {
-    ($fun:expr, $token:expr, $stream:expr, $d:expr) => {{
-        let mut start = $stream.pos() as isize + $d;
-        debug_assert!(start >= 0);
-        if start < 0 { start = 0; }
+    ($fun:expr, $stream:expr, $err:ident) => {{
+        let start = $stream.pos();
         $fun.map_err(|e|
-            Error::InvalidToken($token, $stream.gen_text_pos_from(start as usize), Some(e))
+            Error::$err(e, $stream.gen_text_pos_from(start))
         )
     }}
 }
@@ -415,303 +352,252 @@ impl<'a> Tokenizer<'a> {
     ///
     /// By default, `xmlparser` will check for DTD, root element, etc.
     /// But if we have to parse an XML fragment, it will lead to an error.
-    /// This method switches the parser to the root element content parsing mode.
-    /// So it will treat any data as a content of the root element.
-    pub fn enable_fragment_mode(&mut self) {
-        self.state = State::Elements;
-        self.fragment_parsing = true;
+    /// This method switches the parser to the root element content parsing mode,
+    /// so it will treat any data as a content of the root element.
+    pub fn from_fragment(full_text: &'a str, fragment: core::ops::Range<usize>) -> Self {
+        Tokenizer {
+            stream: Stream::from_substr(full_text, fragment),
+            state: State::Elements,
+            depth: 0,
+            fragment_parsing: true,
+        }
     }
 
-    fn parse_next_impl(s: &mut Stream<'a>, state: State) -> Option<Result<Token<'a>>> {
+    fn parse_next_impl(&mut self) -> Option<Result<Token<'a>>> {
+        let s = &mut self.stream;
+
         if s.at_end() {
             return None;
         }
 
         let start = s.pos();
 
-        if start == 0 {
-            // Skip UTF-8 BOM.
-            if s.starts_with(&[0xEF, 0xBB, 0xBF]) {
-                s.advance(3);
-            }
-        }
-
-        macro_rules! parse_token_type {
-            () => ({
-                match Self::parse_token_type(s, state) {
-                    Ok(v) => v,
-                    Err(_) => {
-                        let pos = s.gen_text_pos_from(start);
-                        return Some(Err(Error::UnknownToken(pos)));
-                    }
-                }
-            })
-        }
-
-        macro_rules! gen_err {
-            ($token_type:expr) => ({
-                let pos = s.gen_text_pos_from(start);
-                if $token_type == TokenType::Unknown {
-                    return Some(Err(Error::UnknownToken(pos)));
+        match self.state {
+            State::Declaration => {
+                self.state = State::AfterDeclaration;
+                if s.starts_with(b"<?xml ") {
+                    Some(Self::parse_declaration(s))
                 } else {
-                    return Some(Err(Error::UnexpectedToken($token_type, pos)));
+                    None
                 }
-            })
-        }
+            }
+            State::AfterDeclaration => {
+                if s.starts_with(b"<!DOCTYPE") {
+                    let t = Self::parse_doctype(s);
+                    match t {
+                        Ok(Token::DtdStart { .. }) => self.state = State::Dtd,
+                        Ok(Token::EmptyDtd { .. }) => self.state = State::AfterDtd,
+                        _ => {}
+                    }
 
-        let t = match state {
-            State::Start => {
-                let token_type = parse_token_type!();
-                match token_type {
-                    TokenType::XMLDecl => {
-                        // XML declaration allowed only at the start of the document.
-                        if start == 0 {
-                            Self::parse_declaration(s)
-                        } else {
-                            gen_err!(token_type);
-                        }
-                    }
-                    TokenType::Comment => {
-                        Self::parse_comment(s)
-                    }
-                    TokenType::PI => {
-                        Self::parse_pi(s)
-                    }
-                    TokenType::DoctypeDecl => {
-                        Self::parse_doctype(s)
-                    }
-                    TokenType::ElementStart => {
-                        Self::parse_element_start(s)
-                    }
-                    TokenType::Whitespace => {
-                        s.skip_spaces();
-                        return Self::parse_next_impl(s, state);
-                    }
-                    _ => {
-                        gen_err!(token_type);
+                    Some(t)
+                } else if s.starts_with(b"<!--") {
+                    Some(Self::parse_comment(s))
+                } else if s.starts_with(b"<?") {
+                    if s.starts_with(b"<?xml ") {
+                        Some(Err(Error::UnknownToken(s.gen_text_pos())))
+                    } else {
+                        Some(Self::parse_pi(s))
                     }
+                } else if s.starts_with_space() {
+                    s.skip_spaces();
+                    None
+                } else {
+                    self.state = State::AfterDtd;
+                    None
                 }
             }
             State::Dtd => {
-                let token_type = parse_token_type!();
-                match token_type {
-                      TokenType::ElementDecl
-                    | TokenType::NotationDecl
-                    | TokenType::AttlistDecl => {
-                        if Self::consume_decl(s).is_err() {
-                            gen_err!(token_type);
+                if s.starts_with(b"<!ENTITY") {
+                    Some(Self::parse_entity_decl(s))
+                } else if s.starts_with(b"<!--") {
+                    Some(Self::parse_comment(s))
+                } else if s.starts_with(b"<?") {
+                    if s.starts_with(b"<?xml ") {
+                        Some(Err(Error::UnknownToken(s.gen_text_pos())))
+                    } else {
+                        Some(Self::parse_pi(s))
+                    }
+                } else if s.starts_with(b"]") {
+                    // DTD ends with ']' S? '>', therefore we have to skip possible spaces.
+                    s.advance(1);
+                    s.skip_spaces();
+                    match s.curr_byte() {
+                        Ok(b'>') => {
+                            self.state = State::AfterDtd;
+                            s.advance(1);
+                            Some(Ok(Token::DtdEnd { span: s.slice_back(start) }))
+                        }
+                        Ok(c) => {
+                            let e = StreamError::InvalidChar(c, b'>', s.gen_text_pos());
+                            Some(Err(Error::InvalidDoctype(e, s.gen_text_pos_from(start))))
+                        }
+                        Err(_) => {
+                            let e = StreamError::UnexpectedEndOfStream;
+                            Some(Err(Error::InvalidDoctype(e, s.gen_text_pos_from(start))))
                         }
-
-                        return Self::parse_next_impl(s, state);
                     }
-                    TokenType::EntityDecl => {
-                        Self::parse_entity_decl(s)
-                    }
-                    TokenType::Comment => {
-                        Self::parse_comment(s)
-                    }
-                    TokenType::PI => {
-                        Self::parse_pi(s)
-                    }
-                    TokenType::DoctypeEnd => {
-                        Ok(Token::DtdEnd { span: s.slice_back(s.pos() - 2) })
-                    }
-                    TokenType::Whitespace => {
-                        s.skip_spaces();
-                        return Self::parse_next_impl(s, state);
-                    }
-                    _ => {
-                        gen_err!(token_type);
+                } else if s.starts_with_space() {
+                    s.skip_spaces();
+                    None
+                } else if    s.starts_with(b"<!ELEMENT")
+                          || s.starts_with(b"<!ATTLIST")
+                          || s.starts_with(b"<!NOTATION")
+                {
+                    if Self::consume_decl(s).is_err() {
+                        let pos = s.gen_text_pos_from(start);
+                        Some(Err(Error::UnknownToken(pos)))
+                    } else {
+                        None
                     }
+                } else {
+                    Some(Err(Error::UnknownToken(s.gen_text_pos())))
                 }
             }
             State::AfterDtd => {
-                let token_type = parse_token_type!();
-                match token_type {
-                    TokenType::Comment => {
-                        Self::parse_comment(s)
-                    }
-                    TokenType::PI => {
-                        Self::parse_pi(s)
-                    }
-                    TokenType::ElementStart => {
-                        Self::parse_element_start(s)
-                    }
-                    TokenType::Whitespace => {
-                        s.skip_spaces();
-                        return Self::parse_next_impl(s, state);
-                    }
-                    _ => {
-                        gen_err!(token_type);
+                if s.starts_with(b"<!--") {
+                    Some(Self::parse_comment(s))
+                } else if s.starts_with(b"<?") {
+                    if s.starts_with(b"<?xml ") {
+                        Some(Err(Error::UnknownToken(s.gen_text_pos())))
+                    } else {
+                        Some(Self::parse_pi(s))
                     }
+                } else if s.starts_with(b"<!") {
+                    Some(Err(Error::UnknownToken(s.gen_text_pos())))
+                } else if s.starts_with(b"<") {
+                    self.state = State::Attributes;
+                    Some(Self::parse_element_start(s))
+                } else if s.starts_with_space() {
+                    s.skip_spaces();
+                    None
+                } else {
+                    Some(Err(Error::UnknownToken(s.gen_text_pos())))
                 }
             }
             State::Elements => {
-                let token_type = parse_token_type!();
-                match token_type {
-                    TokenType::ElementStart => {
-                        Self::parse_element_start(s)
+                // Use `match` only here, because only this section is performance-critical.
+                match s.curr_byte() {
+                    Ok(b'<') => {
+                        match s.next_byte() {
+                            Ok(b'!') => {
+                                if s.starts_with(b"<!--") {
+                                    Some(Self::parse_comment(s))
+                                } else if s.starts_with(b"<![CDATA[") {
+                                    Some(Self::parse_cdata(s))
+                                } else {
+                                    Some(Err(Error::UnknownToken(s.gen_text_pos())))
+                                }
+                            }
+                            Ok(b'?') => {
+                                if !s.starts_with(b"<?xml ") {
+                                    Some(Self::parse_pi(s))
+                                } else {
+                                    Some(Err(Error::UnknownToken(s.gen_text_pos())))
+                                }
+                            }
+                            Ok(b'/') => {
+                                if self.depth > 0 {
+                                    self.depth -= 1;
+                                }
+
+                                if self.depth == 0 && !self.fragment_parsing {
+                                    self.state = State::AfterElements;
+                                } else {
+                                    self.state = State::Elements;
+                                }
+
+                                Some(Self::parse_close_element(s))
+                            }
+                            Ok(_) => {
+                                self.state = State::Attributes;
+                                Some(Self::parse_element_start(s))
+                            }
+                            Err(_) => {
+                                return Some(Err(Error::UnknownToken(s.gen_text_pos())));
+                            }
+                        }
                     }
-                    TokenType::ElementClose => {
-                        Self::parse_close_element(s)
+                    Ok(_) => {
+                        Some(Self::parse_text(s))
                     }
-                    TokenType::CDSect => {
-                        Self::parse_cdata(s)
-                    }
-                    TokenType::PI => {
-                        Self::parse_pi(s)
-                    }
-                    TokenType::Comment => {
-                        Self::parse_comment(s)
-                    }
-                    TokenType::CharData => {
-                        Self::parse_text(s)
-                    }
-                    _ => {
-                        gen_err!(token_type);
+                    Err(_) => {
+                        Some(Err(Error::UnknownToken(s.gen_text_pos())))
                     }
                 }
             }
             State::Attributes => {
-                Self::parse_attribute(s).map_err(|e|
-                    Error::InvalidToken(TokenType::Attribute,
-                                        s.gen_text_pos_from(start), Some(e)))
+                let t = Self::parse_attribute(s);
+
+                if let Ok(Token::ElementEnd { end, .. }) = t {
+                    if end == ElementEnd::Open {
+                        self.depth += 1;
+                    }
+
+                    if self.depth == 0 && !self.fragment_parsing {
+                        self.state = State::AfterElements;
+                    } else {
+                        self.state = State::Elements;
+                    }
+                }
+
+                Some(t.map_err(|e| Error::InvalidAttribute(e, s.gen_text_pos_from(start))))
             }
             State::AfterElements => {
-                let token_type = parse_token_type!();
-                match token_type {
-                    TokenType::Comment => {
-                        Self::parse_comment(s)
-                    }
-                    TokenType::PI => {
-                        Self::parse_pi(s)
-                    }
-                    TokenType::Whitespace => {
-                        s.skip_spaces();
-                        return Self::parse_next_impl(s, state);
-                    }
-                    _ => {
-                        gen_err!(token_type);
+                if s.starts_with(b"<!--") {
+                    Some(Self::parse_comment(s))
+                } else if s.starts_with(b"<?") {
+                    if s.starts_with(b"<?xml ") {
+                        Some(Err(Error::UnknownToken(s.gen_text_pos())))
+                    } else {
+                        Some(Self::parse_pi(s))
                     }
+                } else if s.starts_with_space() {
+                    s.skip_spaces();
+                    None
+                } else {
+                    Some(Err(Error::UnknownToken(s.gen_text_pos())))
                 }
             }
             State::End => {
-                return None;
+                None
             }
-        };
-
-        Some(t)
-    }
-
-    fn parse_token_type(s: &mut Stream, state: State) -> StreamResult<TokenType> {
-        let c1 = s.curr_byte()?;
-
-        let t = match c1 {
-            b'<' => {
-                s.advance(1);
-
-                let c2 = s.curr_byte()?;
-                match c2 {
-                    b'?' => {
-                        // TODO: technically, we should check for any whitespace
-                        if s.starts_with(b"?xml ") {
-                            s.advance(5);
-                            TokenType::XMLDecl
-                        } else {
-                            s.advance(1);
-                            TokenType::PI
-                        }
-                    }
-                    b'!' => {
-                        s.advance(1);
-
-                        let c3 = s.curr_byte()?;
-                        match c3 {
-                            b'-' if s.starts_with(b"--") => {
-                                s.advance(2);
-                                TokenType::Comment
-                            }
-                            b'D' if s.starts_with(b"DOCTYPE") => {
-                                s.advance(7);
-                                TokenType::DoctypeDecl
-                            }
-                            b'E' if s.starts_with(b"ELEMENT") => {
-                                s.advance(7);
-                                TokenType::ElementDecl
-                            }
-                            b'A' if s.starts_with(b"ATTLIST") => {
-                                s.advance(7);
-                                TokenType::AttlistDecl
-                            }
-                            b'E' if s.starts_with(b"ENTITY") => {
-                                s.advance(6);
-                                TokenType::EntityDecl
-                            }
-                            b'N' if s.starts_with(b"NOTATION") => {
-                                s.advance(8);
-                                TokenType::NotationDecl
-                            }
-                            b'[' if s.starts_with(b"[CDATA[") => {
-                                s.advance(7);
-                                TokenType::CDSect
-                            }
-                            _ => {
-                                TokenType::Unknown
-                            }
-                        }
-                    }
-                    b'/' => {
-                        s.advance(1);
-                        TokenType::ElementClose
-                    }
-                    _ => {
-                        TokenType::ElementStart
-                    }
-                }
-            }
-            b']' if state == State::Dtd && s.starts_with(b"]>") => {
-                s.advance(2);
-                TokenType::DoctypeEnd
-            }
-            _ => {
-                match state {
-                    State::Start | State::AfterDtd | State::AfterElements | State::Dtd => {
-                        if s.starts_with_space() {
-                            TokenType::Whitespace
-                        } else {
-                            TokenType::Unknown
-                        }
-                    }
-                    State::Elements => {
-                        TokenType::CharData
-                    }
-                    _ => {
-                        TokenType::Unknown
-                    }
-                }
-            }
-        };
-
-        Ok(t)
+        }
     }
 
     fn parse_declaration(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        map_err_at!(Self::parse_declaration_impl(s), TokenType::XMLDecl, s, -6)
+        map_err_at!(Self::parse_declaration_impl(s), s, InvalidDeclaration)
     }
 
     // XMLDecl ::= '<?xml' VersionInfo EncodingDecl? SDDecl? S? '?>'
     fn parse_declaration_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
-        let start = s.pos() - 6;
+        fn consume_spaces(s: &mut Stream) -> StreamResult<()> {
+            if s.starts_with_space() {
+                s.skip_spaces();
+            } else if !s.starts_with(b"?>") && !s.at_end() {
+                return Err(StreamError::InvalidSpace(s.curr_byte_unchecked(), s.gen_text_pos()));
+            }
+
+            Ok(())
+        }
+
+        let start = s.pos();
+        s.advance(6);
 
         let version = Self::parse_version_info(s)?;
+        consume_spaces(s)?;
+
         let encoding = Self::parse_encoding_decl(s)?;
+        if encoding.is_some() {
+            consume_spaces(s)?;
+        }
+
         let standalone = Self::parse_standalone(s)?;
 
         s.skip_spaces();
         s.skip_string(b"?>")?;
 
         let span = s.slice_back(start);
-
         Ok(Token::Declaration { version, encoding, standalone, span })
     }
 
@@ -733,16 +619,14 @@ impl<'a> Tokenizer<'a> {
         Ok(ver)
     }
 
-
     // EncodingDecl ::= S 'encoding' Eq ('"' EncName '"' | "'" EncName "'" )
     // EncName      ::= [A-Za-z] ([A-Za-z0-9._] | '-')*
     fn parse_encoding_decl(s: &mut Stream<'a>) -> StreamResult<Option<StrSpan<'a>>> {
-        s.skip_spaces();
-
-        if s.skip_string(b"encoding").is_err() {
+        if !s.starts_with(b"encoding") {
             return Ok(None);
         }
 
+        s.advance(8);
         s.consume_eq()?;
         let quote = s.consume_quote()?;
         // [A-Za-z] ([A-Za-z0-9._] | '-')*
@@ -761,12 +645,11 @@ impl<'a> Tokenizer<'a> {
 
     // SDDecl ::= S 'standalone' Eq (("'" ('yes' | 'no') "'") | ('"' ('yes' | 'no') '"'))
     fn parse_standalone(s: &mut Stream<'a>) -> StreamResult<Option<bool>> {
-        s.skip_spaces();
-
-        if s.skip_string(b"standalone").is_err() {
+        if !s.starts_with(b"standalone") {
             return Ok(None);
         }
 
+        s.advance(10);
         s.consume_eq()?;
         let quote = s.consume_quote()?;
 
@@ -789,27 +672,24 @@ impl<'a> Tokenizer<'a> {
     }
 
     fn parse_comment(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        let start = s.pos() - 4;
+        let start = s.pos();
         Self::parse_comment_impl(s)
-            .map_err(|_| Error::InvalidToken(TokenType::Comment, s.gen_text_pos_from(start), None))
+            .map_err(|e| Error::InvalidComment(e, s.gen_text_pos_from(start)))
     }
 
     // '<!--' ((Char - '-') | ('-' (Char - '-')))* '-->'
     fn parse_comment_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
-        let start = s.pos() - 4;
-
-        let text = s.consume_chars(|s, c| {
-            if c == '-' && s.starts_with(b"-->") {
-                return false;
-            }
-
-            c.is_xml_char()
-        });
-
+        let start = s.pos();
+        s.advance(4);
+        let text = s.consume_chars(|s, c| !(c == '-' && s.starts_with(b"-->")))?;
         s.skip_string(b"-->")?;
 
         if text.as_str().contains("--") {
-            return Err(StreamError::UnexpectedEndOfStream); // Error type doesn't matter.
+            return Err(StreamError::InvalidCommentData);
+        }
+
+        if text.as_str().ends_with('-') {
+            return Err(StreamError::InvalidCommentEnd);
         }
 
         let span = s.slice_back(start);
@@ -818,30 +698,17 @@ impl<'a> Tokenizer<'a> {
     }
 
     fn parse_pi(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        map_err_at!(Self::parse_pi_impl(s), TokenType::PI, s, -2)
+        map_err_at!(Self::parse_pi_impl(s), s, InvalidPI)
     }
 
     // PI       ::= '<?' PITarget (S (Char* - (Char* '?>' Char*)))? '?>'
     // PITarget ::= Name - (('X' | 'x') ('M' | 'm') ('L' | 'l'))
     fn parse_pi_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
-        let start = s.pos() - 2;
-
+        let start = s.pos();
+        s.advance(2);
         let target = s.consume_name()?;
-
         s.skip_spaces();
-
-        let content = s.consume_chars(|s, c| {
-            if c == '?' && s.starts_with(b"?>") {
-                return false;
-            }
-
-            if !c.is_xml_char() {
-                return false;
-            }
-
-            true
-        });
-
+        let content = s.consume_chars(|s, c| !(c == '?' && s.starts_with(b"?>")))?;
         let content = if !content.is_empty() {
             Some(content)
         } else {
@@ -856,12 +723,13 @@ impl<'a> Tokenizer<'a> {
     }
 
     fn parse_doctype(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        map_err_at!(Self::parse_doctype_impl(s), TokenType::DoctypeDecl, s, -9)
+        map_err_at!(Self::parse_doctype_impl(s), s, InvalidDoctype)
     }
 
     // doctypedecl ::= '<!DOCTYPE' S Name (S ExternalID)? S? ('[' intSubset ']' S?)? '>'
     fn parse_doctype_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
-        let start = s.pos() - 9;
+        let start = s.pos();
+        s.advance(9);
 
         s.consume_spaces()?;
         let name = s.consume_name()?;
@@ -918,14 +786,15 @@ impl<'a> Tokenizer<'a> {
     }
 
     fn parse_entity_decl(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        map_err_at!(Self::parse_entity_decl_impl(s), TokenType::EntityDecl, s, -8)
+        map_err_at!(Self::parse_entity_decl_impl(s), s, InvalidEntity)
     }
 
     // EntityDecl  ::= GEDecl | PEDecl
     // GEDecl      ::= '<!ENTITY' S Name S EntityDef S? '>'
     // PEDecl      ::= '<!ENTITY' S '%' S Name S PEDef S? '>'
     fn parse_entity_decl_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
-        let start = s.pos() - 8;
+        let start = s.pos();
+        s.advance(8);
 
         s.consume_spaces()?;
 
@@ -989,14 +858,13 @@ impl<'a> Tokenizer<'a> {
     }
 
     fn consume_decl(s: &mut Stream) -> StreamResult<()> {
-        s.consume_spaces()?;
         s.skip_bytes(|_, c| c != b'>');
         s.consume_byte(b'>')?;
         Ok(())
     }
 
     fn parse_cdata(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        map_err_at!(Self::parse_cdata_impl(s), TokenType::CDSect, s, -9)
+        map_err_at!(Self::parse_cdata_impl(s), s, InvalidCdata)
     }
 
     // CDSect  ::= CDStart CData CDEnd
@@ -1004,26 +872,22 @@ impl<'a> Tokenizer<'a> {
     // CData   ::= (Char* - (Char* ']]>' Char*))
     // CDEnd   ::= ']]>'
     fn parse_cdata_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
-        let start = s.pos() - 9;
-
-        let text = s.consume_bytes(|s, c| {
-            !(c == b']' && s.starts_with(b"]]>"))
-        });
-
+        let start = s.pos();
+        s.advance(9);
+        let text = s.consume_chars(|s, c| !(c == ']' && s.starts_with(b"]]>")))?;
         s.skip_string(b"]]>")?;
-
         let span = s.slice_back(start);
-
         Ok(Token::Cdata { text, span })
     }
 
     fn parse_element_start(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        map_err_at!(Self::parse_element_start_impl(s), TokenType::ElementStart, s, -1)
+        map_err_at!(Self::parse_element_start_impl(s), s, InvalidElement)
     }
 
     // '<' Name (S Attribute)* S? '>'
     fn parse_element_start_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
-        let start = s.pos() - 1;
+        let start = s.pos();
+        s.advance(1);
         let (prefix, local) = s.consume_qname()?;
         let span = s.slice_back(start);
 
@@ -1031,12 +895,13 @@ impl<'a> Tokenizer<'a> {
     }
 
     fn parse_close_element(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        map_err_at!(Self::parse_close_element_impl(s), TokenType::ElementClose, s, -2)
+        map_err_at!(Self::parse_close_element_impl(s), s, InvalidElement)
     }
 
     // '</' Name S? '>'
     fn parse_close_element_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
-        let start = s.pos() - 2;
+        let start = s.pos();
+        s.advance(2);
 
         let (prefix, tag_name) = s.consume_qname()?;
         s.skip_spaces();
@@ -1049,6 +914,8 @@ impl<'a> Tokenizer<'a> {
 
     // Name Eq AttValue
     fn parse_attribute(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
+        let attr_start = s.pos();
+        let has_space = s.starts_with_space();
         s.skip_spaces();
 
         if let Ok(c) = s.curr_byte() {
@@ -1070,23 +937,47 @@ impl<'a> Tokenizer<'a> {
             }
         }
 
+        if !has_space {
+            if !s.at_end() {
+                return Err(StreamError::InvalidSpace(
+                    s.curr_byte_unchecked(), s.gen_text_pos_from(attr_start))
+                );
+            } else {
+                return Err(StreamError::UnexpectedEndOfStream);
+            }
+        }
+
         let start = s.pos();
 
         let (prefix, local) = s.consume_qname()?;
         s.consume_eq()?;
         let quote = s.consume_quote()?;
+        let quote_c = quote as char;
         // The attribute value must not contain the < character.
-        let value = s.consume_bytes(|_, c| c != quote && c != b'<');
+        let value = s.consume_chars(|_, c| c != quote_c && c != '<')?;
         s.consume_byte(quote)?;
         let span = s.slice_back(start);
 
-        s.skip_spaces();
-
         Ok(Token::Attribute { prefix, local, value, span })
     }
 
     fn parse_text(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        let text = s.consume_bytes(|_, c| c != b'<');
+        map_err_at!(Self::parse_text_impl(s), s, InvalidCharData)
+    }
+
+    fn parse_text_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
+        let text = s.consume_chars(|_, c| c != '<')?;
+
+        // According to the spec, `]]>` must not appear inside a Text node.
+        // https://www.w3.org/TR/xml/#syntax
+        //
+        // Search for `>` first, since it's a bit faster than looking for `]]>`.
+        if text.as_str().contains('>') {
+            if text.as_str().contains("]]>") {
+                return Err(StreamError::InvalidCharacterData);
+            }
+        }
+
         Ok(Token::Text { text })
     }
 }
@@ -1096,44 +987,14 @@ impl<'a> Iterator for Tokenizer<'a> {
 
     #[inline]
     fn next(&mut self) -> Option<Self::Item> {
-        if self.stream.at_end() || self.state == State::End {
-            return None;
+        let mut t = None;
+        while !self.stream.at_end() && self.state != State::End && t.is_none() {
+            t = self.parse_next_impl();
         }
 
-        let t = Self::parse_next_impl(&mut self.stream, self.state);
-
-        if let Some(ref t) = t {
-            match *t {
-                Ok(t) => match t {
-                    Token::ElementStart { .. } => {
-                        self.state = State::Attributes;
-                    }
-                    Token::ElementEnd { ref end, .. } => {
-                        match *end {
-                            ElementEnd::Open => self.depth += 1,
-                            ElementEnd::Close(..) if self.depth > 0 => self.depth -= 1,
-                            _ => {}
-                        }
-
-                        if self.depth == 0 && !self.fragment_parsing {
-                            self.state = State::AfterElements;
-                        } else {
-                            self.state = State::Elements;
-                        }
-                    }
-                    Token::DtdStart { .. } => {
-                        self.state = State::Dtd;
-                    }
-                    Token::EmptyDtd { .. } | Token::DtdEnd { .. } => {
-                        self.state = State::AfterDtd;
-                    }
-                    _ => {}
-                },
-                Err(_) => {
-                    self.stream.jump_to_end();
-                    self.state = State::End;
-                }
-            }
+        if let Some(Err(_)) = t {
+            self.stream.jump_to_end();
+            self.state = State::End;
         }
 
         t
